"""EcoWitt server tests."""

import pytest

from aioecowitt import server

from .const import GW2000A_DATA

# pylint: disable=redefined-outer-name


@pytest.fixture
def ecowitt_server():
    """EcoWitt server fixture."""
    ecowitt_server = server.EcoWittListener()
    yield ecowitt_server


@pytest.fixture
def ecowitt_http(event_loop, aiohttp_raw_server, aiohttp_client, ecowitt_server):
    """EcoWitt HTTP fixture."""
    raw_server = event_loop.run_until_complete(
        aiohttp_raw_server(ecowitt_server.handler)
    )
    return event_loop.run_until_complete(aiohttp_client(raw_server))


@pytest.mark.asyncio
async def test_server_start(ecowitt_server, ecowitt_http) -> None:
    """Test server start."""
    sensors = []

    def on_change(sensor: server.EcoWittSensor) -> None:
        """Test callback."""
        sensors.append(sensor)

    ecowitt_server.new_sensor_cb.append(on_change)

    resp = await ecowitt_http.post("/", data=GW2000A_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    assert len(sensors) == 52


@pytest.mark.asyncio
async def test_server_token(ecowitt_server, ecowitt_http) -> None:
    """Test server start."""
    sensors = []
    path = "/test"
    ecowitt_server.path = path

    def on_change(sensor: server.EcoWittSensor) -> None:
        """Test callback."""
        sensors.append(sensor)

    ecowitt_server.new_sensor_cb.append(on_change)

    resp = await ecowitt_http.post("/", data=GW2000A_DATA)
    assert resp.status == 404

    resp = await ecowitt_http.post(path, data=GW2000A_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    assert len(sensors) == 52



"""Weather datapoint calculator."""
from __future__ import annotations

import meteocalc


def _ftoc(fahrenheit: float | str) -> float:
    """Convert f to c."""
    return round(meteocalc.Temp(fahrenheit, "F").c, 1)


def _volt_to_percent(volt: float, low: float, high: float) -> float | int:
    percent = round(((volt - low) / (high - low)) * 100)
    if percent < 0:
        percent = 0
    elif percent > 100:
        percent = 100
    return percent


def weather_datapoints(data: dict[str, str]) -> dict[str, str | int | float | None]:
    """Calculate and convert weather data."""
    mph_kmh = 1.60934
    in_hpa = 33.86
    in_mm = 25.4
    km_mi = 0.6213712
    wm_lux = 0.0079

    # basic conversions
    if "humidityin" in data:
        data["humidityin"] = int(data["humidityin"])
    if "humidity" in data:
        data["humidity"] = int(data["humidity"])
    if "winddir" in data:
        data["winddir"] = int(data["winddir"])
    if "winddir_avg10m" in data:
        data["winddir_avg10m"] = int(data["winddir_avg10m"])
    if "uv" in data:
        data["uv"] = int(data["uv"])
    if "solarradiation" in data:
        data["solarradiation"] = float(data["solarradiation"])
        data["solarradiation_lux"] = round(data["solarradiation"] / wm_lux, 1)

    # lightning
    if "lightning_time" in data:
        if data["lightning_time"] is not None and data["lightning_time"] != "":
            data["lightning_time"] = int(data["lightning_time"])
    if "lightning_num" in data:
        data["lightning_num"] = int(data["lightning_num"])
    if "lightning" in data:
        if data["lightning"] is not None and data["lightning"] != "":
            data["lightning"] = int(data["lightning"])
            data["lightning_mi"] = int(round(data["lightning"] * km_mi))

    # temperatures
    if "tempf" in data:
        data["tempf"] = float(data["tempf"])
        data["tempc"] = _ftoc(data["tempf"])
    if "tempinf" in data:
        data["tempinf"] = float(data["tempinf"])
        data["tempinc"] = _ftoc(data["tempinf"])
    # (WH45)
    if "tf_co2" in data:
        data["tf_co2"] = float(data["tf_co2"])
        data["tf_co2c"] = _ftoc(data["tf_co2"])
    # WN34 Soil Temperature Sensor
    for j in range(1, 9):
        wnf = f"tf_ch{j}"
        wnc = f"tf_ch{j}c"
        if wnf in data:
            data[wnf] = float(data[wnf])
            data[wnc] = _ftoc(data[wnf])

    # numbered WH31 temp/humid
    for j in range(1, 9):
        tmpf = f"temp{j}f"
        tmpc = f"temp{j}c"
        hum = f"humidity{j}"
        if tmpf in data:
            data[tmpf] = float(data[tmpf])
            data[tmpc] = _ftoc(data[tmpf])
        if hum in data:
            data[hum] = int(data[hum])

    # speeds
    if "windspeedmph" in data:
        data["windspeedmph"] = float(data["windspeedmph"])
        data["windspeedkmh"] = round(data["windspeedmph"] * mph_kmh, 1)
    if "windgustmph" in data:
        data["windgustmph"] = float(data["windgustmph"])
        data["windgustkmh"] = round(data["windgustmph"] * mph_kmh, 1)
    # I assume this is MPH?
    if "maxdailygust" in data:
        data["maxdailygust"] = float(data["maxdailygust"])
        data["maxdailygustkmh"] = round(data["maxdailygust"] * mph_kmh, 1)
    if "windspdmph_avg10m" in data:
        data["windspdmph_avg10m"] = float(data["windspdmph_avg10m"])
        data["windspdkmh_avg10m"] = round(float(data["windspdmph_avg10m"] * mph_kmh), 1)

    # distances
    if "rainratein" in data:
        data["rainratein"] = float(data["rainratein"])
        data["rainratemm"] = round(data["rainratein"] * in_mm, 1)
    if "eventrainin" in data:
        data["eventrainin"] = float(data["eventrainin"])
        data["eventrainmm"] = round(data["eventrainin"] * in_mm, 1)
    if "hourlyrainin" in data:
        data["hourlyrainin"] = float(data["hourlyrainin"])
        data["hourlyrainmm"] = round(data["hourlyrainin"] * in_mm, 1)
    if "dailyrainin" in data:
        data["dailyrainin"] = float(data["dailyrainin"])
        data["dailyrainmm"] = round(data["dailyrainin"] * in_mm, 1)
    if "weeklyrainin" in data:
        data["weeklyrainin"] = float(data["weeklyrainin"])
        data["weeklyrainmm"] = round(data["weeklyrainin"] * in_mm, 1)
    if "monthlyrainin" in data:
        data["monthlyrainin"] = float(data["monthlyrainin"])
        data["monthlyrainmm"] = round(data["monthlyrainin"] * in_mm, 1)
    if "yearlyrainin" in data:
        data["yearlyrainin"] = float(data["yearlyrainin"])
        data["yearlyrainmm"] = round(data["yearlyrainin"] * in_mm, 1)
    if "totalrainin" in data:
        data["totalrainin"] = float(data["totalrainin"])
        data["totalrainmm"] = round(data["totalrainin"] * in_mm, 1)

    # piezo rain sensor
    if "rrain_piezo" in data:
        data["rrain_piezo"] = float(data["rrain_piezo"])
        data["rrain_piezomm"] = round(data["rrain_piezo"] * in_mm, 1)
    if "erain_piezo" in data:
        data["erain_piezo"] = float(data["erain_piezo"])
        data["erain_piezomm"] = round(data["erain_piezo"] * in_mm, 1)
    if "hrain_piezo" in data:
        data["hrain_piezo"] = float(data["hrain_piezo"])
        data["hrain_piezomm"] = round(data["hrain_piezo"] * in_mm, 1)
    if "drain_piezo" in data:
        data["drain_piezo"] = float(data["drain_piezo"])
        data["drain_piezomm"] = round(data["drain_piezo"] * in_mm, 1)
    if "wrain_piezo" in data:
        data["wrain_piezo"] = float(data["wrain_piezo"])
        data["wrain_piezomm"] = round(data["wrain_piezo"] * in_mm, 1)
    if "mrain_piezo" in data:
        data["mrain_piezo"] = float(data["mrain_piezo"])
        data["mrain_piezomm"] = round(data["mrain_piezo"] * in_mm, 1)
    if "yrain_piezo" in data:
        data["yrain_piezo"] = float(data["yrain_piezo"])
        data["yrain_piezomm"] = round(data["yrain_piezo"] * in_mm, 1)

    # Pressure
    if "baromrelin" in data:
        data["baromrelin"] = float(data["baromrelin"])
        data["baromrelhpa"] = round(data["baromrelin"] * in_hpa, 1)
    if "baromabsin" in data:
        data["baromabsin"] = float(data["baromabsin"])
        data["baromabshpa"] = round(data["baromabsin"] * in_hpa, 1)

    # Wind chill
    if "tempf" in data and "windspeedmph" in data:
        try:
            wind_chill = meteocalc.wind_chill(data["tempf"], data["windspeedmph"])
        except ValueError:
            data["windchillf"] = None
            data["windchillc"] = None
        else:
            data["windchillf"] = round(wind_chill.f, 1)
            data["windchillc"] = round(wind_chill.c, 1)

    # Dew point
    for j in ["", "in", "1", "2", "3", "4", "5", "6", "7", "8"]:
        if f"temp{j}c" in data and f"humidity{j}" in data:
            dewpoint = meteocalc.dew_point(data[f"temp{j}c"], data[f"humidity{j}"])
            data[f"dewpoint{j}c"] = round(dewpoint.c, 1)
            data[f"dewpoint{j}f"] = round(dewpoint.f, 1)

    # Feels like
    if "tempf" in data and "humidity" in data and "windspeedmph" in data:
        feels_like = meteocalc.feels_like(
            data["tempf"], data["windspeedmph"], data["humidity"]
        )
        data["tempfeelsf"] = round(feels_like.f, 1)
        data["tempfeelsc"] = round(feels_like.c, 1)

    # Soil moisture (WH51)
    for j in range(1, 9):
        name = f"soilmoisture{j}"
        if name in data:
            data[name] = int(data[name])

    # PM 2.5 sensor (WH41)
    for j in range(1, 5):
        pmc = f"pm25_ch{j}"
        pma = f"pm25_avg_24h_ch{j}"
        if pmc in data:
            data[pmc] = float(data[pmc])
        if pma in data:
            data[pma] = float(data[pma])

    # Leak sensor (WH55)
    for j in range(1, 5):
        leak = f"leak_ch{j}"
        if leak in data:
            data[leak] = int(data[leak])

    # CO2 indoor air quality (WH45) (note temp is in temps above)
    pm_floats = [
        "pm25",
        "pm25_24h",
        "pm10",
        "pm10_24",
    ]
    for prefix in pm_floats:
        name = f"{prefix}_co2"
        if name in data:
            data[name] = float(data[name])
    if "co2" in data:
        data["co2"] = int(data["co2"])
    if "co2_24h" in data:
        data["co2_24h"] = int(data["co2_24h"])
    if "humi_co2" in data:
        data["humi_co2"] = int(data["humi_co2"])

    # Batteries
    bat_names = [
        "wh25",
        "wh26",
        "wh40",
        "wh57",
        "wh65",
        "wh68",
        "wh80",
        "wh90",
        "co2_",
    ]
    bat_range_names = [
        "soil",
        "",  # for just 'batt'
        "pm25",
        "leak",
        "tf_",  # WN34 voltage type
    ]

    for prefix in bat_names:
        name = f"{prefix}batt"
        if name in data:
            data[name] = float(data[name])

    for r_prefix in bat_range_names:
        for j in range(1, 9):
            name = f"{r_prefix}batt{j}"
            if name in data:
                data[name] = float(data[name])

    # percentage battery for device view
    if "wh90batt" in data:
        data["wh90battpc"] = _volt_to_percent(data["wh90batt"], 2.4, 3.0)
        data["ws90cap_volt"] = float(data["ws90cap_volt"])

    return data



"""Test ecowitt sensor module."""

from aioecowitt import sensor


def test_update_listener() -> None:
    """Test on change get updates from callback."""
    ecowit_sensor = sensor.EcoWittSensor(
        "test", "test", sensor.EcoWittSensorTypes.TEMPERATURE_C, "test"
    )

    called = False

    def on_change() -> None:
        """Test callback."""
        nonlocal called
        called = True

    ecowit_sensor.update_cb.append(on_change)

    ecowit_sensor.update_value(10, 0, 0)
    assert called

    called = False
    ecowit_sensor.update_value(10, 0, 0)
    assert not called

    ecowit_sensor.update_value(11, 0, 0)
    assert called



"""aioEcoWitt API server."""
from __future__ import annotations

import logging
import time
from typing import Callable

from aiohttp import web

from .sensor import EcoWittSensor, SENSOR_MAP, EcoWittSensorTypes
from . import calc

_LOGGER = logging.getLogger(__name__)
_ECOWITT_LISTEN_PORT = 49199


class EcoWittListener:
    """EcoWitt Server API server."""

    def __init__(self, port: int = _ECOWITT_LISTEN_PORT, path: str = None):
        """Initialize EcoWitt Server."""
        # API Constants
        self.port: int = port
        self.path: str = path

        # webserver
        self.server: None | web.Server = None
        self.runner: None | web.ServerRunner = None
        self.site: None | web.TCPSite = None

        # internal data
        self.last_values: dict[str, str | int | float | None] = {}
        self.lastupd: float = 0
        self.new_sensor_cb: list[Callable[[EcoWittSensor], None]] = []

        # storage
        self._station_type: str = "Unknown"
        self._station_freq: str = "Unknown"
        self._station_model: str = "Unknown"
        self._mac_addr: None | str = None

        self.sensors: dict[str, EcoWittSensor] = {}

    def _new_sensor_cb(self, sensor: EcoWittSensor) -> None:
        """Internal new sensor callback

        binds to self.new_sensor_cb
        """
        for callback in self.new_sensor_cb:
            callback(sensor)

    def process_data(self, weather_data: dict[str, str | float | int | None]) -> None:
        """Process data from weather station."""
        station = weather_data["PASSKEY"]
        last_update = time.time()
        last_update_m = time.monotonic()

        for datapoint in weather_data.keys():
            sensor = self.sensors.get(datapoint)
            if sensor is None:
                # we have a new sensor
                if datapoint not in SENSOR_MAP:
                    _LOGGER.warning(
                        "Unhandled sensor type %s value %s",
                        datapoint,
                        weather_data[datapoint],
                    )
                    continue
                metadata = SENSOR_MAP[datapoint]

                sensor = EcoWittSensor(
                    metadata.name, datapoint, metadata.stype, station
                )
                self.sensors[datapoint] = sensor
                try:
                    self._new_sensor_cb(sensor)
                except Exception as err:  # pylint: disable=broad-except
                    _LOGGER.warning("EcoWitt new sensor callback error: %s", err)

            try:
                sensor.update_value(weather_data[datapoint], last_update, last_update_m)
            except Exception as err:  # pylint: disable=broad-except
                _LOGGER.warning("Sensor update error: %s", err)

    async def handler(self, request: web.BaseRequest) -> web.Response:
        """AIOHTTP handler for the API."""
        if request.method != "POST":
            return web.Response(status=405)
        if self.path is not None and request.path != self.path:
            return web.Response(status=404)
        data = await request.post()

        # data is not a dict, it's a MultiDict
        weather_data = calc.weather_datapoints({**data})
        self.last_values = weather_data.copy()
        self.lastupd = time.time()
        self.process_data(weather_data)

        return web.Response(text="OK")

    async def start(self) -> None:
        """Listen and process."""

        self.server = web.Server(self.handler)
        self.runner = web.ServerRunner(self.server)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, port=self.port)
        await self.site.start()

    async def stop(self) -> None:
        """Stop listening."""
        await self.site.stop()

    def list_sensor_keys(self) -> list[str]:
        """List all available sensors by key."""
        return [sensor.key for sensor in self.sensors]

    def list_sensor_keys_by_type(self, stype: EcoWittSensorTypes) -> list[str]:
        """List all available sensors of a given type."""
        sensor_list = []
        for sensor in self.sensors:
            if sensor.stype == stype:
                sensor_list.append(sensor.key)
        return sensor_list

    def get_sensor_value_by_key(self, key: str) -> None | str | int | float:
        """Find the sensor named key and return its value."""
        sensor = self.sensors.get(key)
        if sensor is None:
            return None
        return sensor.value



"""Setup module for EcoWitt."""
from pathlib import Path

from setuptools import find_packages, setup

PROJECT_DIR = Path(__file__).parent.resolve()
README_FILE = PROJECT_DIR / "README.md"
VERSION = "2022.7.0"


setup(
    name="aioecowitt",
    version=VERSION,
    url="https://github.com/home-assistant-libs/aioecowitt",
    download_url="https://github.com/home-assistant-libs/aioecowitt",
    author="Home Assistant Team",
    author_email="hello@home-assistant.io",
    description="Python wrapper for EcoWitt Protocol",
    long_description=README_FILE.read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["test.*", "test", "misc"]),
    package_data={"aioecowitt": ["py.typed"]},
    python_requires=">=3.9",
    install_requires=["aiohttp>3", "meteocalc>=1.1.0"],
    entry_points={"console_scripts": ["ecowitt-testserver = aioecowitt.__main__:main"]},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Home Automation",
    ],
)
